@@ -2,7 +2,7 @@
 
     Compatibility name file for  devfsd  (build compatibility names).
 
-    Copyright (C) 1998-2000  Richard Gooch
+    Copyright (C) 1998-2001  Richard Gooch
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -30,9 +30,11 @@
 
     Written by      Richard Gooch   30-APR-2000: Copied from devfsd.c
 
-    Last updated by Richard Gooch   10-JUN-2000: Added compatibility entries
+    Updated by      Richard Gooch   10-JUN-2000: Added compatibility entries
   for ISDN BRI driver. Thanks to Xuan Baldauf <xuan--lkml@baldauf.org>.
 
+    Last updated by Chris Rankin    22-JAN-2001: Added entries for I2C devices.
+
 
 */
 #include <linux/major.h>
@@ -85,6 +87,7 @@ static struct translate_struct translate_table[] =
     {"vc/",        "tty%s"},
     {"misc/",      NULL},
     {"isdn/",      NULL},
+    {"i2c/",       "i2c-%s"},
     {NULL,         NULL}
 };
 
