@@ -165,13 +165,31 @@
     Updated by      Richard Gooch   3-JUL-2000: Added "-C /etc/modules.devfs"
   when calling modprobe(8). Fail if a configuration line has EXECUTE modprobe.
 
-    Last updated by Richard Gooch   4-JUL-2000: Added #define __USE_GNU. Thanks
+    Updated by      Richard Gooch   4-JUL-2000: Added #define __USE_GNU. Thanks
   to Adam J. Richter <adam@yggdrasil.com>.
 
-    Last updated by Richard Gooch   5-FEB-2001: Switched from __USE_GNU to
+    Updated by      Chris Rankin    21-JAN-2001: Allow the daemon to substitute
+  regular expressions, similarly to sed. E.g. the expression ^foo\([0-9]\)$ has
+  a subexpression within the \(, \) markers. The string foo4 matches this
+  expression: the tag \0 would be replaced by "foo4", its first subexpression
+  \1 by "4", etc.
+
+    Updated by      Richard Gooch   5-FEB-2001: Switched from __USE_GNU to
   _GNU_SOURCE, according to recommendation from Ulrich. Do dummy opens of
   /dev/null so that fds [0:2] are open (and thus may be safely closed later).
 
+    Updated by      Richard Gooch   21-FEB-2001: Bug fix in function
+  <action_call_function>: a failed MFUNCTION would result in a bogus CFUNCTION
+  being called.
+
+    Updated by      Richard Gooch   27-JUL-2001: Bug fixes when execvp(3) fails
+  (child did not exit).
+
+    Last updated by Richard Gooch   30-JUL-2001: Applied and cleaned up patch
+  by Chris Rankin (regular subexpression support). Fixed potential buffer
+  overrun problems by using <snprintf>. Thanks to Sebastian Krahmer
+  <krahmer@suse.de>. Added "-np" switch. Thanks to John Fremlin.
+
 
 */
 #define _GNU_SOURCE
@@ -210,6 +228,7 @@
 
 #define CONFIG_FILE "/etc/devfsd.conf"
 #define MAX_ARGS (6 + 1)
+#define MAX_SUBEXPR  10
 
 /*  Update only after changing code to reflect new protocol  */
 #define DEVFSD_PROTOCOL_REVISION_DAEMON  5
@@ -329,11 +348,15 @@ static void action_permissions (const struct devfsd_notify_struct *info,
 static void action_modload (const struct devfsd_notify_struct *info,
 			    const struct config_entry_struct *entry);
 static void action_execute (const struct devfsd_notify_struct *info,
-			    const struct config_entry_struct *entry);
+			    const struct config_entry_struct *entry,
+			    const regmatch_t *regexpr, unsigned int numexpr);
 static void action_call_function (const struct devfsd_notify_struct *info,
-				  const struct config_entry_struct *entry);
+				  const struct config_entry_struct *entry,
+				  const regmatch_t *regexpr,
+				  unsigned int numexpr);
 static void action_copy (const struct devfsd_notify_struct *info,
-			 const struct config_entry_struct *entry);
+			 const struct config_entry_struct *entry,
+			 const regmatch_t *regexpr, unsigned int number);
 static void action_compat (const struct devfsd_notify_struct *info,
 			   unsigned int action);
 static void free_config ();
@@ -348,6 +371,18 @@ static void do_scan_and_service (DIR *dp, CONST char *dirname);
 static flag make_symlink (const char *oldpath, const char *newpath);
 static flag make_dir_tree (const char *path);
 static struct shared_object *get_shared_object (CONST char *name);
+static flag expand_expression(char *output, unsigned int outsize,
+			      const char *input,
+			      const char *(*get_variable)(const char *variable,
+                                                          void *info),
+                              void *info,
+                              const char *devname,
+                              const regmatch_t *ex, unsigned int numexp,
+                              FILE *errfp);
+static void expand_regexp (char *output, unsigned int outsize,
+			   const char *input, const char *devname,
+			   const regmatch_t *ex, unsigned int numex,
+			   FILE *errfp);
 
 
 /*  Private data  */
@@ -379,12 +414,13 @@ int main (int argc, char **argv)
     flag print_version = FALSE;
     flag debug_devfs = FALSE;
     flag do_daemon = TRUE;
+    flag no_polling = FALSE;
     int fd, proto_rev, count;
     unsigned int devfs_debug_mask;
     unsigned long event_mask = 0;
     struct sigaction new_action;
     struct stat statbuf;
-    static char usage[] = "devfsd mntpnt [-v] [-d] [-t num] [-D mask] [-fg]";
+    static char usage[] = "devfsd mntpnt [-v] [-d] [-t num] [-D mask] [-fg] [-np]";
 
     if (argc < 2)
     {
@@ -407,6 +443,7 @@ int main (int argc, char **argv)
 	    debug_devfs = TRUE;
 	}
 	else if (strcmp (argv[count], "-fg") == 0) do_daemon = FALSE;
+	else if (strcmp (argv[count], "-np") == 0) no_polling = TRUE;
 	else
 	{
 	    fprintf (stderr, "Usage:\t%s\n", usage);
@@ -505,6 +542,7 @@ int main (int argc, char **argv)
     /*  Do the scan before forking, so that boot scripts see the finished
 	product  */
     do_scan_and_service (opendir (mount_point), mount_point);
+    if (no_polling) exit (0);
     if (do_daemon)
     {
 	/*  Release so that the child can grab it  */
@@ -869,7 +907,8 @@ static void service_name (const struct devfsd_notify_struct *info)
     [RETURNS] Nothing.
 */
 {
-    regmatch_t mbuf;
+    unsigned int n;
+    regmatch_t mbuf[MAX_SUBEXPR];
     struct config_entry_struct *entry;
 
     if (info->overrun_count > 0)
@@ -889,7 +928,9 @@ static void service_name (const struct devfsd_notify_struct *info)
 	/*  First check if action matches the type, then check if name matches
 	 */
 	if (info->type != entry->action.when) continue;
-	if (regexec (&entry->preg, info->devname, 1, &mbuf, 0) != 0) continue;
+	if (regexec (&entry->preg, info->devname, MAX_SUBEXPR, mbuf, 0) != 0)
+	    continue;
+	for (n = 0; (n < MAX_SUBEXPR) && (mbuf[n].rm_so != -1); ++n);
 	switch (entry->action.what)
 	{
 	  case AC_PERMISSIONS:
@@ -899,14 +940,14 @@ static void service_name (const struct devfsd_notify_struct *info)
 	    action_modload (info, entry);
 	    break;
 	  case AC_EXECUTE:
-	    action_execute (info, entry);
+	    action_execute (info, entry, mbuf, n);
 	    break;
 	  case AC_MFUNCTION:
 	  case AC_CFUNCTION:
-	    action_call_function (info, entry);
+	    action_call_function (info, entry, mbuf, n);
 	    break;
 	  case AC_COPY:
-	    action_copy (info, entry);
+	    action_copy (info, entry, mbuf, n);
 	    break;
 	  case AC_IGNORE:
 	    return;
@@ -938,7 +979,7 @@ static void action_permissions (const struct devfsd_notify_struct *info,
 
     if (lstat (info->devname, &statbuf) != 0)
     {
-	SYSLOG (LOG_ERR, "error stat(2)ing: \"%s\"\t%s\n",
+	SYSLOG (LOG_ERR, "error lstat(2)ing: \"%s\"\t%s\n",
 		info->devname, ERRSTRING);
 	SYSLOG (LOG_ERR, "exiting\n");
 	exit (1);
@@ -1007,8 +1048,7 @@ static void action_modload (const struct devfsd_notify_struct *info,
     argv[3] = "/etc/modules.devfs";
     argv[4] = device;
     argv[5] = NULL;
-    strcpy (device, "/dev/");
-    strcpy (device + 5, info->devname);
+    snprintf (device, sizeof (device), "/dev/%s", info->devname);
     if (modprobe != NULL)
     {
 	(*modprobe) (5, argv);
@@ -1032,13 +1072,18 @@ static void action_modload (const struct devfsd_notify_struct *info,
     }
     execvp (argv[0], argv);
     SYSLOG (LOG_ERR, "error execing: \"%s\"\t%s\n", argv[0], ERRSTRING);
+    _exit (1);
 }   /*  End Function action_modload  */
 
 static void action_execute (const struct devfsd_notify_struct *info,
-			    const struct config_entry_struct *entry)
+			    const struct config_entry_struct *entry,
+			    const regmatch_t *regexpr, unsigned int numexpr)
 /*  [SUMMARY] Execute a programme.
     <info> The devfs change.
     <entry> The config file entry.
+    <regexpr> The number of subexpression (start, end) offsets within the
+    device name.
+    <numexpr> The number of elements within <<regexpr>>.
     [RETURNS] Nothing.
 */
 {
@@ -1049,17 +1094,17 @@ static void action_execute (const struct devfsd_notify_struct *info,
 
     gv_info.info = info;
     gv_info.devname = info->devname;
-    strcpy (gv_info.devpath, mount_point);
-    strcat (gv_info.devpath, "/");
-    strcat (gv_info.devpath, info->devname);
+    snprintf (gv_info.devpath, sizeof (gv_info.devpath), "%s/%s",
+	      mount_point, info->devname);
     if (trace_level > 1) fprintf (stderr, "Calling %s for \"%s\" | \"%s\":",
 				  entry->u.execute.argv[0], info->devname,
 				  gv_info.devpath);
     for (count = 0; entry->u.execute.argv[count] != NULL; ++count)
     {
-	st_expr_expand (largv[count], STRING_LENGTH,
-			entry->u.execute.argv[count], get_variable, &gv_info,
-			NULL);
+	expand_expression (largv[count], STRING_LENGTH,
+			   entry->u.execute.argv[count],
+			   get_variable, &gv_info,
+			   gv_info.devname, regexpr, numexpr, NULL);
 	argv[count] = largv[count];
 	if (trace_level > 1) fprintf (stderr, " '%s'", argv[count]);
     }
@@ -1084,13 +1129,19 @@ static void action_execute (const struct devfsd_notify_struct *info,
     }
     execvp (argv[0], argv);
     SYSLOG (LOG_ERR, "error execing: \"%s\"\t%s\n", argv[0], ERRSTRING);
+    _exit (1);
 }   /*  End Function action_execute  */
 
 static void action_call_function (const struct devfsd_notify_struct *info,
-				  const struct config_entry_struct *entry)
+				  const struct config_entry_struct *entry,
+				  const regmatch_t *regexpr,
+				  unsigned int numexpr)
 /*  [SUMMARY] Call a function.
     <info> The devfs change.
     <entry> The config file entry.
+    <regexpr> The number of subexpression (start, end) offsets within the
+    device name.
+    <numexpr> The number of elements within <<regexpr>>.
     [RETURNS] Nothing.
 */
 {
@@ -1101,9 +1152,8 @@ static void action_call_function (const struct devfsd_notify_struct *info,
 
     gv_info.info = info;
     gv_info.devname = info->devname;
-    strcpy (gv_info.devpath, mount_point);
-    strcat (gv_info.devpath, "/");
-    strcat (gv_info.devpath, info->devname);
+    snprintf (gv_info.devpath, sizeof (gv_info.devpath), "%s/%s",
+	      mount_point, info->devname);
     if (trace_level > 1)
 	fprintf (stderr, "Calling %s in %s for \"%s\" | \"%s\":",
 		 entry->u.function.argv[0], entry->u.function.so->name,
@@ -1115,9 +1165,10 @@ static void action_call_function (const struct devfsd_notify_struct *info,
 	    argv[count] = (char *) info;
 	else
 	{
-	    st_expr_expand (largv[count], STRING_LENGTH,
-			    entry->u.function.argv[count], get_variable,
-			    &gv_info, NULL);
+	    expand_expression (largv[count], STRING_LENGTH,
+			       entry->u.function.argv[count], get_variable,
+			       &gv_info, gv_info.devname, regexpr, numexpr,
+			       NULL);
 	    argv[count] = largv[count];
 	    if (trace_level > 1) fprintf (stderr, " '%s'", argv[count]);
 	}
@@ -1130,17 +1181,24 @@ static void action_call_function (const struct devfsd_notify_struct *info,
 	if ( (*entry->u.function.func.m) (count, argv) == 0 )
 	    return;
     }
-    if ( (*entry->u.function.func.c) (argv[1], argv[2], argv[3], argv[4],
-				      argv[5]) == 0 ) return;
+    else
+    {
+	if ( (*entry->u.function.func.c) (argv[1], argv[2], argv[3], argv[4],
+					  argv[5]) == 0 ) return;
+    }
     SYSLOG (LOG_ERR, "error calling: \"%s\" in \"%s\"\n",
 	    entry->u.function.argv[0], argv[0]);
 }   /*  End Function action_call_function  */
 
 static void action_copy (const struct devfsd_notify_struct *info,
-			 const struct config_entry_struct *entry)
+			 const struct config_entry_struct *entry,
+                         const regmatch_t *regexpr, unsigned int numexpr)
 /*  [SUMMARY] Copy permissions.
     <info> The devfs change.
     <entry> The config file entry.
+    <regexpr> This list of subexpression (start, end) offsets within the
+    device name.
+    <numexpr> The number of elements in <<regexpr>>.  
     [RETURNS] Nothing.
 */
 {
@@ -1150,13 +1208,14 @@ static void action_copy (const struct devfsd_notify_struct *info,
 
     gv_info.info = info;
     gv_info.devname = info->devname;
-    strcpy (gv_info.devpath, mount_point);
-    strcat (gv_info.devpath, "/");
-    strcat (gv_info.devpath, info->devname);
-    st_expr_expand (source, STRING_LENGTH, entry->u.copy.source, get_variable,
-		    &gv_info, NULL);
-    st_expr_expand (destination, STRING_LENGTH, entry->u.copy.destination,
-		    get_variable, &gv_info, NULL);
+    snprintf (gv_info.devpath, sizeof (gv_info.devpath), "%s/%s",
+	      mount_point, info->devname);
+    expand_expression (source, STRING_LENGTH, entry->u.copy.source,
+		       get_variable, &gv_info, gv_info.devname,
+		       regexpr, numexpr, NULL);
+    expand_expression (destination, STRING_LENGTH, entry->u.copy.destination,
+		       get_variable, &gv_info, gv_info.devname,
+		       regexpr, numexpr, NULL);
     if (trace_level > 1)
 	fprintf (stderr, "Copying %s to %s for \"%s\" | \"%s\"\n",
 		 source, destination, info->devname, gv_info.devpath);
@@ -1205,7 +1264,8 @@ static void action_copy (const struct devfsd_notify_struct *info,
       case S_IFSOCK:
 	if ( ( fd = socket (AF_UNIX, SOCK_STREAM, 0) ) < 0 ) break;
 	un_addr.sun_family = AF_UNIX;
-	strcpy (un_addr.sun_path, destination);
+	snprintf (un_addr.sun_path, sizeof (un_addr.sun_path), "%s",
+		  destination);
 	if (bind (fd, (struct sockaddr *) &un_addr, (int) sizeof un_addr) != 0)
 	    break;
 	if (chmod (destination, source_stat.st_mode & ~S_IFMT) != 0) break;
@@ -1274,7 +1334,7 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    sscanf (info->devname + 5, "host%d/bus%d/target%d/lun%d/",
 		    &host, &bus, &target, &lun);
 	    compat_name = compat_buf;
-	    sprintf (dest_buf, "../%s", info->devname);
+	    snprintf (dest_buf, sizeof (dest_buf), "../%s", info->devname);
 	    dest_name = dest_buf;
 	    if (strncmp (ptr = (strrchr (info->devname, '/') + 1), "mt", 2)
 		== 0)
@@ -1322,7 +1382,7 @@ static void action_compat (const struct devfsd_notify_struct *info,
 	    sscanf (info->devname + 4, "host%d/bus%d/target%d/lun%d/",
 		    &host, &bus, &target, &lun);
 	    compat_name = compat_buf;
-	    sprintf (dest_buf, "../%s", info->devname + 4);
+	    snprintf (dest_buf, sizeof (dest_buf), "../%s", info->devname + 4);
 	    dest_name = dest_buf;
 	    if (strcmp (info->devname + info->namelen - 4, "disc") == 0)
 		sprintf (compat_buf, "ide/hd/c%db%dt%du%d",
@@ -1336,8 +1396,9 @@ static void action_compat (const struct devfsd_notify_struct *info,
 			 host, bus, target,lun);
 	    else if (strncmp (ptr = (strrchr (info->devname, '/') + 1), "mt",
 			      2) == 0)
-		sprintf (compat_buf, "ide/mt/c%db%dt%du%d%s",
-			 host, bus, target, lun, ptr + 2);
+		snprintf (compat_buf, sizeof (compat_buf),
+			  "ide/mt/c%db%dt%du%d%s",
+			  host, bus, target, lun, ptr + 2);
 	    else compat_name = NULL;
 	}
 	break;
@@ -1617,10 +1678,10 @@ static void do_scan_and_service (DIR *dp, CONST char *dirname)
 
 	if ( (strcmp (de->d_name, ".") == 0) ||
 	     (strcmp (de->d_name, "..") == 0) ) continue;
-	sprintf (path, "%s/%s", dirname, de->d_name);
+	snprintf (path, sizeof (path), "%s/%s", dirname, de->d_name);
 	if (lstat (path, &statbuf) != 0)
 	{
-	    SYSLOG (LOG_ERR, "Error stat(2)ing file \"%s\"\t%s\n",
+	    SYSLOG (LOG_ERR, "Error lstat(2)ing file \"%s\"\t%s\n",
 		    path, ERRSTRING);
 	    exit (RV_SYS_ERROR);
 	}
@@ -1636,7 +1697,8 @@ static void do_scan_and_service (DIR *dp, CONST char *dirname)
 	info.minor = MINOR (statbuf.st_rdev);
 	info.uid = statbuf.st_uid;
 	info.gid = statbuf.st_gid;
-	strcpy (info.devname, path + strlen (mount_point) + 1);
+	snprintf (info.devname, sizeof (info.devname), "%s",
+		  path + strlen (mount_point) + 1);
 	info.namelen = strlen (info.devname);
 	service_name (&info);
     }
@@ -1679,7 +1741,7 @@ static flag make_dir_tree (const char *path)
     char path_copy[STRING_LENGTH];
     /*static char function_name[] = "make_dir_tree";*/
 
-    strcpy (path_copy, path);
+    snprintf (path_copy, sizeof (path_copy), "%s", path);
     ptr1 = path_copy;
     while (TRUE)
     {
@@ -1733,3 +1795,96 @@ static struct shared_object *get_shared_object (CONST char *name)
     first_so = so;
     return (so);
 }   /*  End Function get_shared_object  */
+
+static flag expand_expression(char *output, unsigned int outsize,
+			      const char *input,
+			      const char *(*get_variable)(const char *variable,
+                                                          void *info),
+                              void *info,
+                              const char *devname,
+                              const regmatch_t *ex, unsigned int numexp,
+                              FILE *errfp)
+/*  [SUMMARY] Expand enviroment variables and regular subexpressions in string.
+    <output> The output expanded expression is written here.
+    <length> The size of the output buffer.
+    <input> The input expression. This may equal <<output>>.
+    <get_variable> A function which will be used to get variable values. If
+    this returns NULL, the environment is searched instead. If this is NULL,
+    only the environment is searched.
+    <info> An arbitrary pointer passed to <<get_variable>>.
+    <devname> Device name; specifically, this is the string that contains all
+    of the regular subexpressions.
+    <ex> Array of start / end offsets into info->devname for each subexpression
+    <numexp> Number of regular subexpressions found in <<devname>>.
+    <errfp> Diagnostic messages are written here. If this is NULL the global
+    stderr is used instead.
+    [RETURNS] TRUE on success, else FALSE.
+*/
+{
+    char temp[STRING_LENGTH];
+
+    if ( !st_expr_expand (temp, STRING_LENGTH, input,
+			  get_variable, info, errfp) ) return (FALSE);
+    expand_regexp (output, outsize, temp, devname, ex, numexp, errfp);
+    return (TRUE);
+}   /*  End Function expand_expression  */
+
+static void expand_regexp (char *output, size_t outsize, const char *input,
+			   const char *devname,
+			   const regmatch_t *ex, unsigned int numex,
+			   FILE *errfp)
+/*  [SUMMARY] Expand all occurrences of the regular subexpressions \0 to \9.
+    <output> The output expanded expression is written here.
+    <outsize> The size of the output buffer.
+    <input> The input expression. This may NOT equal <<output>>, because
+    supporting that would require yet another string-copy. However, it's not
+    hard to write a simple wrapper function to add this functionality for those
+    few cases that need it.
+    <devname> Device name; specifically, this is the string that contains all
+    of the regular subexpressions.
+    <ex> An array of start and end offsets into <<devname>>, one for each
+    subexpression
+    <numex> Number of subexpressions in the offset-array <<ex>>.
+    <errfp> Diagnostic messages are written here. If this is NULL the global
+    stderr is used instead.
+*/
+{
+    const char last_exp = '0' - 1 + numex;
+    int c = -1;
+
+    if (!errfp) errfp = stderr;
+    /*  Guarantee NULL termination by writing an explicit '\0' character into
+	the very last byte  */
+    if (outsize) output[--outsize] = '\0';
+    /*  Copy the input string into the output buffer, replacing '\\' with '\'
+	and '\0' .. '\9' with subexpressions 0 .. 9, if they exist. Other \x
+	codes are deleted  */
+    while ( (c != '\0') && (outsize != 0) )
+    {
+        c = *input;
+        ++input;
+        if (c == '\\')
+        {
+            c = *input;
+            ++input;
+            if (c != '\\')
+            {
+                if ((c >= '0') && (c <= last_exp))
+                {
+                    const regmatch_t *subexp = ex + (c - '0');
+                    unsigned int sublen = subexp->rm_eo - subexp->rm_so;
+
+                    /*  Range checking  */
+                    if (sublen > outsize) sublen = outsize;
+                    strncpy (output, devname + subexp->rm_so, sublen);
+                    output += sublen;
+                    outsize -= sublen;
+                }
+                continue;
+            }
+        }
+        *output = c;
+        ++output;
+        --outsize;
+    } /* while */
+}   /*  End Function expand_regexp  */
